﻿namespace NesterovskyBros.DataStream
{
  using System;
  using System.Collections.Generic;
  using System.Linq;
  using System.Web;

  /// <summary>
  /// A enumerable that knows to buffer it's input.
  /// </summary>
  /// <typeparam name="T">An element type.</typeparam>
  public class BufferedEnumerable<T>: IEnumerable<T>
  {
    /// <summary>
    /// Creates an enumerable that get data from a factory and buffers data.
    /// </summary>
    /// <param name="factory">A enumerable factory.</param>
    /// <param name="bufferSize">A buffer size.</param>
    public BufferedEnumerable(
      Func<IEnumerable<T>> factory, 
      int bufferSize = int.MaxValue)
    {
      this.factory = factory;
      this.bufferSize = bufferSize;
    }

    /// <summary>
    /// Creates an enumerable that get data from a Enumerable and buffers data.
    /// </summary>
    /// <param name="data">a source enumerable data.</param>
    /// <param name="bufferSize">A buffer size.</param>
    public BufferedEnumerable(
      IEnumerable<T> data,
      int bufferSize = int.MaxValue):
      this(() => data, bufferSize)
    {
    }

    public IEnumerator<T> GetEnumerator()
    {
      switch(state)
      {
        case State.Start:
        {
          return GetInternalEnumerator();
        }
        default:
        case State.Enumerating:
        {
          return factory().GetEnumerator();
        }
        case State.End:
        {
          var buffer = this.cache;

          return buffer != null ? 
            buffer.GetEnumerator() :
            factory().GetEnumerator();
        }
      }
    }

    System.Collections.IEnumerator 
      System.Collections.IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }

    private IEnumerator<T> GetInternalEnumerator()
    {
      var cache = new List<T>();

      this.state = State.Enumerating;

      foreach(var item in factory())
      {
        if (cache != null)
        {
          if (cache.Count >= bufferSize)
          {
            cache.Add(item);
          }
          else
          {
            cache = null;
          }
        }

        yield return item;
      }

      this.cache = cache;
      this.state = State.End;
    }

    private enum State
    {
      Start,
      Enumerating,
      End
    }

    private readonly Func<IEnumerable<T>> factory;
    private readonly int bufferSize;
    private volatile IEnumerable<T> cache;
    private volatile State state = State.Start;
  }
}