﻿namespace NesterovskyBros.DataStream
{
  using System;
  using System.Collections.Generic;
  using System.Web;
  using System.IO;

  /// <summary>
  /// A stream reading data from a enumeration of arrays.
  /// </summary>
  public class EnumeratorStream: Stream
  {
    /// <summary>
    /// Creates an EnumeratorStream instance.
    /// </summary>
    /// <param name="source">A source.</param>
    public EnumeratorStream(IEnumerable<byte[]> source)
    {
      this.source = source.GetEnumerator();
    }

    /// <summary>
    /// Disposes a stream.
    /// </summary>
    /// <param name="disposing">Disposing indicator.</param>
    protected override void Dispose(bool disposing)
    {
      try
      {
        if (disposing)
        {
          var source = this.source;

          if (source != null)
          {
            this.source = null;
            this.current = null;
            this.offset = 0;
            source.Dispose();
          }
        }
      }
      finally
      {
        base.Dispose(disposing);
      }
    }

    /// <summary>
    /// Read is supported.
    /// </summary>
    public override bool CanRead
    {
      get { return true; }
    }

    /// <summary>
    /// Write is not supported.
    /// </summary>
    public override bool CanWrite
    {
      get { return false; }
    }

    /// <summary>
    /// No seek is supported.
    /// </summary>
    public override bool CanSeek
    {
      get { return false; }
    }


    /// <summary>
    /// Gets stream position.
    /// </summary>
    public override long Position
    {
      get
      {
        return position;
      }
      set
      {
        throw new NotSupportedException();
      }
    }

    /// <summary>
    /// No Length is supported.
    /// </summary>
    public override long Length
    {
      get { throw new NotSupportedException(); }
    }

    /// <summary>
    /// No SetLength is supported.
    /// </summary>
    public override void SetLength(long value)
    {
      throw new NotSupportedException();
    }

    /// <summary>
    /// No Seek is supported.
    /// </summary>
    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    /// <summary>
    /// Do nothing.
    /// </summary>
    public override void Flush()
    {
    }

    /// <summary>
    /// Reads data from the stream.
    /// </summary>
    /// <param name="buffer">A buffer to put data into.</param>
    /// <param name="offset">A buffer offset.</param>
    /// <param name="count">A count of bytes to read.</param>
    /// <returns>
    /// The total number of bytes read into the buffer. 
    /// This can be less than the number of bytes requested if 
    /// that many bytes are not currently available, or zero (0)
    /// if the end of the stream has been reached.
    /// </returns>
    public override int Read(byte[] buffer, int offset, int count)
    {
      int i = 0;

      while(i < count)
      {
        if (current != null)
        {
          buffer[offset + i++] = current[this.offset++];
          ++position;

          if (this.offset < current.Length)
          {
            continue;
          }

          current = null;
          this.offset = 0;
        }

        var source = this.source;

        if (source == null)
        {
          break;
        }

        if (!source.MoveNext())
        {
          this.source = null;
          source.Dispose();

          break;
        }

        current = source.Current;

        if ((current != null) && (current.Length == 0))
        {
          current = null;
        }
      }

      return i;
    }

    /// <summary>
    /// Write is not supported.
    /// </summary>
    public override void Write(byte[] buffer, int offset, int count)
    {
      throw new NotSupportedException();
    }

    /// <summary>
    /// Source enumerator.
    /// </summary>
    private IEnumerator<byte[]> source;

    /// <summary>
    /// Current item.
    /// </summary>
    private byte[] current;

    /// <summary>
    /// Offset in the current item.
    /// </summary>
    private int offset;

    /// <summary>
    /// Current position.
    /// </summary>
    private long position;
  }
}