﻿namespace NesterovskyBros.DataStream
{
  using System;
  using System.Linq;
  using System.Text;
  using System.IO;
  using System.IO.Pipes;
  using System.Web;
  using System.Web.Caching;
  using System.Xml;
  using System.Xml.Xsl;
  using System.Xml.XPath;
  using System.Xml.Schema;
  using System.Xml.Serialization;
  using System.Collections.Generic;
  using System.Threading.Tasks;
  using System.Runtime.Serialization;

  using Ionic.Zip;

  using NesterovskyBros.Xml;
  using NesterovskyBros.Xsl;

  public static class Extensions
  {
    /// <summary>
    /// A dictionary getter.
    /// </summary>
    /// <typeparam name="K">A key type.</typeparam>
    /// <typeparam name="V">A value class type.</typeparam>
    /// <param name="dictionary">A dictionary to get value from.</param>
    /// <param name="key">A key to get value for.</param>
    /// <returns>
    /// Value when key is not null and it's contained in the dictionary, 
    /// and null otherwise.
    /// </returns>
    public static V Get<K, V>(this IDictionary<K, V> dictionary, K key)
      where V: class
    {
      V value;

      if (key == null)
      {
        value = null;
      }
      else
      {
        dictionary.TryGetValue(key, out value);
      }

      return value;
    }

    /// <summary>
    /// A dictionary getter.
    /// </summary>
    /// <typeparam name="K">A key type.</typeparam>
    /// <typeparam name="V">A value class type.</typeparam>
    /// <param name="dictionary">A dictionary to get value from.</param>
    /// <param name="key">A key to get value for.</param>
    /// <returns>
    /// Value when key is not null and it's contained in the dictionary, 
    /// and null otherwise.
    /// </returns>
    public static V Get<K, V>(this IDictionary<K, V> dictionary, K? key)
      where V: class
      where K: struct
    {
      V value;

      if (key == null)
      {
        value = null;
      }
      else
      {
        dictionary.TryGetValue(key.GetValueOrDefault(), out value);
      }

      return value;
    }

    /// <summary>
    /// Gets a stream joined from source streams.
    /// </summary>
    /// <param name="streams">A source streams to read from.</param>
    /// <param name="closeStreams">
    /// Indicates whether to close streams, which has been reached 
    /// during a enumeration.
    /// </param>
    /// <returns>A joined stream.</returns>
    public static Stream JoinStreams(
      this IEnumerable<Stream> streams, 
      bool closeStreams = true)
    {
      return new JoinedStream(streams, closeStreams);
    }

    /// <summary>
    /// Converts a enumeration of items into input xml stream.
    /// DataContractSerializer is used to serialize items.
    /// </summary>
    /// <typeparam name="T">An item type.</typeparam>
    /// <param name="items">Item to get xml stream from.</param>
    /// <param name="rootName">Root name.</param>
    /// <param name="rootNamespace">Root namespace.</param>
    /// <returns>Xml stream.</returns>
    public static Stream ToXmlStream<T>(
      this IEnumerable<T> items, 
      string rootName = null, 
      string rootNamespace = null)
    {
      return items.ToXmlStreamParts<T>(rootName, rootNamespace).
        JoinStreams(false);
    }

    /// <summary>
    /// Converts a enumeration of items into input xml stream.
    /// DataContractSerializer is used to serialize items.
    /// </summary>
    /// <typeparam name="T">An item type.</typeparam>
    /// <param name="items">Item to get xml stream from.</param>
    /// <param name="rootName">Root name.</param>
    /// <param name="rootNamespace">Root namespace.</param>
    /// <returns>A enumeration of xml stream part.</returns>
    private static IEnumerable<Stream> ToXmlStreamParts<T>(
      this IEnumerable<T> items,
      string rootName = null,
      string rootNamespace = null)
    {
      if (rootName == null)
      {
        rootName = "ArrayOfItems";
      }

      if (rootNamespace == null)
      {
        rootNamespace = "";
      }

      var serializer = new DataContractSerializer(typeof(T));
      var stream = new MemoryStream();
      var writer = XmlDictionaryWriter.CreateTextWriter(stream);

      writer.WriteStartDocument();
      writer.WriteStartElement(rootName, rootNamespace);
      writer.WriteXmlnsAttribute("s", XmlSchema.Namespace);
      writer.WriteXmlnsAttribute("i", XmlSchema.InstanceNamespace);

      foreach(var item in items)
      {
        serializer.WriteObject(writer, item);
        writer.WriteString(" "); // This is to prevent lazy write in xml writer.

        writer.Flush();
        stream.Position = 0;

        yield return stream;

        stream.Position = 0;
        stream.SetLength(0);
      }

      writer.WriteEndElement();
      writer.WriteEndDocument();
      
      writer.Flush();
      stream.Position = 0;

      yield return stream;
    }

    /// <summary>
    /// Returns input stream whose content is generated.
    /// </summary>
    /// <param name="generator">A content generator function.</param>
    /// <returns>An input stream.</returns>
    public static Stream GetPipedStream(Action<Stream> generator)
    {
      var output = new AnonymousPipeServerStream();
      var input = new AnonymousPipeClientStream(
        output.GetClientHandleAsString());

      Task.Factory.StartNew(
        () =>
        {
          using (output)
          {
            generator(output);
            output.WaitForPipeDrain();
          }
        },
        TaskCreationOptions.LongRunning);

      return input;
    }

    /// <summary>
    /// Creates and writes an XSLX file into the specified zip output stream.
    /// </summary>
    /// <param name="stream">the output stream.</param>
    /// <param name="basePath">a base path for template files to be archived.</param>
    /// <param name="factory">
    /// A factory that provides an input stream by its name.
    /// </param>
    /// <param name="args">
    /// A name-value map of XSLT parameters, if any.
    /// </param>
    public static void CreateXslxFile(
      Stream stream,
      string basePath,
      Func<string, Stream> factory = null,
      IDictionary<string, object> args = null)
    {
      // creates a zip archive
      using (var zip = new ZipFile())
      {
        var basePathLength = basePath.Length + 1;
        var files =
          Directory.GetFiles(basePath, "*.*", SearchOption.AllDirectories);

        foreach (var path in files)
        {
          var file = path.Substring(basePathLength);

          switch(Path.GetExtension(file))
          {
            case ".scc":
            case ".bak":
            {
              // exclude these files
              continue;
            }
            case ".xslt":
            {
              // transform and add a zip entry asynchronously
              zip.AddEntry(
                Path.ChangeExtension(file, "xml"),
                (entryName, target) =>
                {
                  // load a cached XSLT file
                  var xslt = Path.Combine(
                    basePath,
                    Path.ChangeExtension(entryName, "xslt"));
                  var transform = GetStylesheet(xslt);

                  // populate arguments, if any
                  var argumentList = new XsltArgumentList();

                  if (args != null)
                  {
                    foreach (var arg in args)
                    {
                      if (arg.Value == null)
                      {
                        continue;
                      }

                      argumentList.AddParam(
                        arg.Key, 
                        "urn:nesterovsky-bros:args", 
                        arg.Value);
                    }
                  }

                  // add extensions functions
                  argumentList.AddExtensionObject(
                    "urn:nesterovsky-bros:fn",
                    new XsltFunctions(factory));

                  var source = XmlReader.Create(new StringReader("<data/>"));

                  // generate XML content
                  transform.Transform(
                    new ForwardXPathNavigator(source),
                    argumentList,
                    XmlWriter.Create(target),
                    new XmlUrlResolver());
                });

              break;
            }
            default:
            {
              // copy a file into zip archive as is
              zip.AddEntry(
                file,
                (entryName, target) =>
                {
                  var filePath = Path.Combine(basePath, entryName);

                  using (var source = File.OpenRead(filePath))
                  {
                    byte[] buffer = new byte[1024];
                    int length = 0;

                    while ((length = source.Read(buffer, 0, 1024)) > 0)
                    {
                      target.Write(buffer, 0, length);
                    }
                  }
                });

              break;
            }
          }
        }

        zip.Save(stream);
      }
    }

    /// <summary>
    /// Gets cached XsltCompiledTransform for the path.
    /// </summary>
    /// <param name="path">A stylesheet path.</param>
    /// <returns>A XsltCompiledTransform instance.</returns>
    public static XslCompiledTransform GetStylesheet(string path)
    {
      path = Path.IsPathRooted(path) ? path :
        HttpContext.Current.Server.MapPath(path);

      var cache = HttpRuntime.Cache;
      var key = "xslt:" + path;
      var transform = cache[key] as XslCompiledTransform;

      if (transform == null)
      {
        transform = new XslCompiledTransform();
        transform.Load(path, XsltSettings.TrustedXslt, new XmlUrlResolver());

        cache.Add(
          key,
          transform,
          new CacheDependency(path),
          Cache.NoAbsoluteExpiration,
          Cache.NoSlidingExpiration,
          CacheItemPriority.Normal,
          null);
      }

      return transform;
    }

    /// <summary>
    /// Creates an enumerable that get data from a Enumerable and buffers data.
    /// </summary>
    /// <param name="data">a source enumerable data.</param>
    /// <param name="bufferSize">optional, defines buffer size in records.</param>
    /// <returns>a buffered enumerable data.</returns>
    public static IEnumerable<T> ToBufferedEnumerable<T>(
      this IEnumerable<T> data,
      int bufferSize = int.MaxValue)
    {
      return new BufferedEnumerable<T>(data, bufferSize);
    }

    /// <summary>
    /// Creates an enumerable that get data from a Enumerable and buffers data.
    /// </summary>
    /// <param name="factory">A enumerable factory.</param>
    /// <param name="bufferSize">optional, defines buffer size in records.</param>
    /// <returns>a buffered enumerable data.</returns>
    public static IEnumerable<T> ToBufferedEnumerable<T>(
      this Func<IEnumerable<T>> factory,
      int bufferSize = int.MaxValue)
    {
      return new BufferedEnumerable<T>(factory, bufferSize);
    }

    /// <summary>
    /// A stream joined from a enumeration of source streams.
    /// </summary>
    private class JoinedStream: Stream
    {
      /// <summary>
      /// Creates JoinedStream instance.
      /// </summary>
      /// <param name="streams">Source streams.</param>
      /// <param name="closeStreams">
      /// Indicates whether to close acquired streams.
      /// </param>
      public JoinedStream(IEnumerable<Stream> streams, bool closeStreams = true)
      {
        this.enumerator = streams.GetEnumerator();
        this.closeStreams = closeStreams;
      }

      /// <summary>
      /// Indicates that current stream supports reading.
      /// </summary>
      public override bool CanRead { get { return true; } }

      /// <summary>
      /// Indicates that Seek is not supported.
      /// </summary>
      public override bool CanSeek { get { return false; } }

      /// <summary>
      /// Indicates that write is not supported.
      /// </summary>
      public override bool CanWrite { get { return false; } }
      
      /// <summary>
      /// Length is not supported.
      /// </summary>
      public override long Length { get { throw new NotSupportedException();  } }

      /// <summary>
      /// Position is not supported.
      /// </summary>
      public override long Position { get; set; }

      /// <summary>
      /// Releases resources.
      /// </summary>
      /// <param name="disposing">
      /// true to release both managed and unmanaged resources; 
      /// false to release only unmanaged resources.
      /// </param>
      protected override void Dispose(bool disposing)
      {
        try
        {
          var current = this.current;
          var enumerator = this.enumerator;

          this.current = null;
          this.enumerator = null;

          using(enumerator)
          {
            if (closeStreams && (current != null))
            {
              current.Dispose();
            }
          }
        }
        finally
        {
          base.Dispose(disposing);
        }
      }

      /// <summary>
      /// Does nothing.
      /// </summary>
      public override void Flush()
      { 
      }
      
      /// <summary>
      /// Reads a sequence of bytes from the current
      /// stream and advances the position within the stream by the 
      /// number of bytes read.
      /// </summary>
      /// <param name="buffer">
      /// An array of bytes. When this method returns, the buffer 
      /// contains the specified byte array with the values between 
      /// offset and (offset + count - 1) replaced by the bytes read 
      /// from the current source.
      /// </param>
      /// <param name="offset">
      /// The zero-based byte offset in buffer at which to begin storing the 
      /// data read from the current stream.
      /// </param>
      /// <param name="count">
      /// The maximum number of bytes to be read from the current stream.
      /// </param>
      /// <returns>
      /// The total number of bytes read into the buffer. 
      /// This can be less than the number of bytes requested if that many 
      /// bytes are not currently available, or zero (0) if the end of the 
      /// stream has been reached.
      /// </returns>
      public override int Read(byte[] buffer, int offset, int count)
      {
        var current = this.current;

        while (true)
        {
          if (current != null)
          {
            int read = current.Read(buffer, offset, count);

            if (read > 0)
            {
              return read;
            }

            this.current = null;

            if (closeStreams)
            {
              current.Close();
            }
          }

          var enumerator = this.enumerator;

          if (enumerator != null)
          {
            if (!enumerator.MoveNext())
            {
              this.enumerator = null;
              enumerator.Dispose();

              return 0;
            }

            this.current = current = enumerator.Current;
          }
        }
      }

      /// <summary>
      /// Seek is not supported.
      /// </summary>
      /// <param name="offset">
      /// A byte offset relative to the origin parameter.
      /// </param>
      /// <param name="origin">
      /// A value of type System.IO.SeekOrigin indicating the reference point used
      /// to obtain the new position.
      /// </param>
      /// <returns>The new position within the current stream.</returns>
      public override long Seek(long offset, SeekOrigin origin)
      {
        throw new NotSupportedException();
      }
      
      /// <summary>
      /// Operation is not supported.
      /// </summary>
      /// <param name="value">
      /// The desired length of the current stream in bytes.
      /// </param>
      public override void SetLength(long value)
      {
        throw new NotSupportedException();
      }

      /// <summary>
      /// Write is not supported.
      /// </summary>
      /// <param name="buffer">
      /// An array of bytes. This method copies count bytes from buffer 
      /// to the current stream.
      /// </param>
      /// <param name="offset">
      /// The zero-based byte offset in buffer at which to begin copying 
      /// bytes to the current stream.
      /// </param>
      /// <param name="count">
      /// The number of bytes to be written to the current stream.
      /// </param>
      public override void Write(byte[] buffer, int offset, int count)
      {
        throw new NotSupportedException();
      }

      /// <summary>
      /// Source streams.
      /// </summary>
      private IEnumerator<Stream> enumerator;
      
      /// <summary>
      /// Current stream.
      /// </summary>
      private Stream current;

      /// <summary>
      /// Indicates whether to close current stream.
      /// </summary>
      private bool closeStreams;
    }
  }
}
