//------------------------------------------------------------------------------
// <copyright file = "ForwardXPathNavigator.cs" 
//   company = "Nesterovsky bros"
//   e-mail = "mailto:contact@nesterovsky-bros.com"
//   home = "http://www.nesterovsky-bros.com">
//     
//      Copyright (c) 2002 Nesterovsky bros.  All rights reserved.
//     
//      The use and distribution terms for this software are contained in the file
//      named license.txt, which can be found in the root of this distribution.
//      By using this software in any fashion, you are agreeing to be bound by the
//      terms of this license.
//     
//      You must not remove this notice, or any other, from this software.
//
// </copyright>                                                                
//------------------------------------------------------------------------------

namespace NesterovskyBros.Xml
{
  using System;
  using System.Text;
  using System.Xml;
  using System.Xml.XPath;

  /// <summary>
  ///   Represents an XPathNavigator that provides fast, non-cached, 
  ///   forward-only access to XML data.
  /// </summary>
  /// <remarks>
  ///   <para>
  ///     Such a navigator is used for performing special XSL tranformations
  ///     that use forward-only access to source data.
  ///   </para>
  ///   <para>
  ///     All properties and methods of ForwardXPathNavigator that are not 
  ///     compatible with forward-only access mode throw XPathException.
  ///   </para>
  ///   <seealso cref = "XPathNavigator"/>
  /// </remarks>
  public class ForwardXPathNavigator: XPathNavigator
  {
    private const string readerIsNotSpecified = "Reader is not specified.";
    private const string operationIsNotAvailable = 
      "Operation is not available. Data is out of scope.";
    internal const string NsXmlns = "http://www.w3.org/2000/xmlns/";
    internal const string NsXml = "http://www.w3.org/XML/1998/namespace";
    internal const string Lang = "lang";

    internal class Context
    {
      internal XPathNodeType nodeType;
      internal RootContext root;
      internal ContainerContext parent;
      internal int documentIndex;
      internal Context firstChild;

      public virtual string BaseUri { get { return parent.BaseUri; } }

      public virtual string Prefix { get { return string.Empty; } }

      public virtual string LocalName { get { return string.Empty; } }

      public virtual string NamespaceURI { get { return string.Empty; } }

      public virtual string Name { get { return string.Empty; } }

      public virtual string Value { get { return string.Empty; } }

      public virtual Context NextSibling { get { return root.CreateContext(parent); } }

      public virtual Context FirstChild { get { return null; } }

      public virtual Context FirstAttribute { get { return null; } }

      public virtual Context NextAttribute { get { return null; } }

      public virtual bool HasAttributes { get { return false; } }

      public virtual AttributeContext GetAttribute(string localName, string namespaceURI)
      { 
        return null; 
      }

      public virtual NamespaceContext FirstNamespace { get { return null; } }

      public virtual NamespaceContext NextNamespace { get { return null; } }
      
      public virtual NamespaceContext GetNamespace(string name) { return null; }
      
      public virtual bool IsElement(string nameAtom, string nsAtom) { return false; }
    }

    internal static readonly Context unknownContext = new Context();

    internal class NamedContext: Context
    {
      internal string prefix;
      internal string localName;
      internal string namespaceURI;

      public override string Prefix { get { return prefix; } }

      public override string LocalName { get { return localName; } }

      public override string NamespaceURI { get { return namespaceURI; } }

      public override string Name 
      {
        get { return Prefix != null && Prefix.Length != 0 ? Prefix+":"+LocalName : LocalName; }
      }
    }

    internal class ContainerContext: NamedContext
    {
      internal string baseURI;
      internal AttributeContext firstAttribute;
      internal NamespaceContext firstNamespace;
      internal string value;
      internal int depth;

      // Cache
      internal ElementContext elementCache;
      internal AttributeContext attributeCache;
      internal NamespaceContext namespaceCache;

      public override string BaseUri { get { return baseURI; } }

      public override Context FirstChild
      { 
        get { return firstChild == unknownContext ? root.CreateContext(this) : firstChild; }
      }

      private string ReadElementValue()
      {
        XmlReader reader = root.reader;
        StringBuilder builder = null;
        string value = null;

        do
        {
          switch(reader.NodeType)
          {
            case XmlNodeType.EndElement:
              if (reader.Depth <= depth)
              {
                firstChild = null;
                goto end;
              }
              break;
            case XmlNodeType.EntityReference:
              reader.ResolveEntity();
              break;
            case XmlNodeType.CDATA:
            case XmlNodeType.Text:
            case XmlNodeType.Whitespace:
            case XmlNodeType.SignificantWhitespace:
              if ((object)value == null)
              {
                value = reader.Value;
              }
              else 
              {
                string v = reader.Value;

                if (builder == null)
                {
                  builder = root.builder;
                  if (builder == null)
                  {
                    root.builder = builder = new StringBuilder(value, value.Length+v.Length);
                  }
                  else
                  {
                    builder.Length = 0;
                    builder.EnsureCapacity(value.Length+v.Length);
                    builder.Append(value);
                  }
                }

                builder.Append(v);
              }
              break;
          }
        }
        while(root.Read());

      end:
        return builder != null ? builder.ToString() : 
          (object)value != null ? value : string.Empty;
      }

      public override string Value
      { 
        get
        {
          if (firstChild == unknownContext)
            value = ReadElementValue();

          return value;
        }
      }

      public override NamespaceContext FirstNamespace { get { return firstNamespace; } }

      public override NamespaceContext GetNamespace(string name) 
      { 
        for(NamespaceContext ns = firstNamespace; ns != null; ns = ns.next) 
        {
          if (ns.localName == name) return ns;
        }

        return null; 
      }
    }

    internal class AttributeContext: NamedContext
    {
      internal AttributeContext next;
      internal string value;

      // Cache
      internal AttributeContext attributeCache;

      internal AttributeContext(RootContext root)
      {
        this.nodeType = XPathNodeType.Attribute;
        this.root = root;
      }

      internal void Init(ContainerContext parent, int documentIndex)
      {
        XmlReader reader = root.reader;

        this.parent = parent;
        this.documentIndex = documentIndex;
        this.prefix = reader.Prefix;
        this.localName = reader.LocalName;
        this.namespaceURI = reader.NamespaceURI;
        this.value = reader.Value;
      }

      public override string Value { get { return value; } }

      public override Context NextAttribute { get { return next; } }

      public override Context NextSibling { get { return null; } }
    }

    internal class NamespaceContext: Context
    {
      internal NamespaceContext next;
      internal string localName;
      internal string value;

      // Cache
      internal NamespaceContext namespaceCache;

      internal NamespaceContext(RootContext root)
      {
        this.nodeType = XPathNodeType.Namespace;
        this.root = root;
      }

      internal void Init(ContainerContext parent, int documentIndex)
      {
        XmlReader reader = root.reader;

        this.parent = parent;
        this.documentIndex = documentIndex;
        this.localName = reader.Prefix.Length == 0 ? string.Empty : reader.LocalName;
        this.value = reader.Value;
      }

      internal void Init(NamespaceContext other)
      {
        this.parent = other.parent;
        this.documentIndex = other.documentIndex;
        this.localName = other.localName;
        this.value = other.value;
      }

      public override string LocalName { get { return localName; } }

      public override string Name { get { return localName; } }

      public override string Value { get { return value; } }

      public override NamespaceContext NextNamespace { get { return next; } }

      public override Context NextSibling { get { return null; } }
    }

    internal class ElementContext: ContainerContext
    {
      internal ElementContext(RootContext root, ContainerContext parent)
      {
        this.nodeType = XPathNodeType.Element;
        this.root = root;
        this.parent = parent;
      }

      internal void Init(int documentIndex)
      {
        XmlReader reader = root.reader;

        this.prefix = reader.Prefix;
        this.localName = reader.LocalName;
        this.namespaceURI = reader.NamespaceURI;
        this.baseURI = reader.BaseURI;
        this.documentIndex = documentIndex;
        this.depth = reader.Depth;
        this.attributeCache = parent.firstAttribute != null ?
          parent.firstAttribute.attributeCache : parent.attributeCache;
        this.namespaceCache = parent.firstNamespace != null ?
          parent.firstNamespace.namespaceCache : parent.namespaceCache;
      }
      
      public override Context NextSibling
      { 
        get 
        { 
          if (firstChild != null)
          {
            // We are in the our child context
            XmlReader reader = root.reader;

            while(reader.Read() && reader.Depth>depth)
            {
              // Skip data.
            }
          }

          // Skip the end.
          root.Read();

          return root.CreateContext(parent);
        }
      }

      public override Context FirstAttribute { get { return firstAttribute; } }

      public override Context NextAttribute { get { return null; } }

      public override bool HasAttributes { get { return firstAttribute != null; } }

      public override AttributeContext GetAttribute(string localName, string namespaceURI) 
      {
        for(AttributeContext attr = firstAttribute; attr !=  null; attr = attr.next) 
        {
          if(attr.localName == localName && attr.namespaceURI == namespaceURI) return attr;
        }

        return null;
      }
      
      public override bool IsElement(string nameAtom, string nsAtom)
      { 
        return ((object)localName == (object)nameAtom || nameAtom.Length == 0) && 
          (object)namespaceURI == (object)nsAtom;
      }
    }

    internal class ValueContext: Context
    {
      internal string value;

      internal void Init(ContainerContext parent, int documentIndex)
      {
        this.parent = parent;
        this.documentIndex = documentIndex;
        this.value = root.reader.Value;
      }

      public override string Value { get { return value; } }
    }

    internal class TextContext: ValueContext
    {
      internal TextContext() {}

      internal TextContext(RootContext root)
      {
        this.nodeType = XPathNodeType.Text;
        this.root = root;
      }
    }

    internal class WhitespaceContext: TextContext
    {
      internal WhitespaceContext() {}

      internal WhitespaceContext(RootContext root)
      {
        this.nodeType = XPathNodeType.Whitespace;
        this.root = root;
      }
    }

    internal class SignificantWhitespaceContext: WhitespaceContext
    {
      internal SignificantWhitespaceContext(RootContext root)
      {
        this.nodeType = XPathNodeType.SignificantWhitespace;
        this.root = root;
      }
    }

    internal class CommentContext: ValueContext
    {
      internal CommentContext(RootContext root)
      {
        this.nodeType = XPathNodeType.Comment;
        this.root = root;
      }
    }

    internal class ProcessingInstructionContext: Context
    {
      internal string name;
      internal string value;

      internal ProcessingInstructionContext(RootContext root)
      {
        this.nodeType = XPathNodeType.ProcessingInstruction;
        this.root = root;
      }
        
      internal void Init(ContainerContext parent, int documentIndex)
      {
        XmlReader reader = root.reader;

        this.parent = parent;
        this.documentIndex = documentIndex;
        this.name = reader.Name;
        this.value = reader.Value;
      }

      public override string Name { get { return name; } }
        
      public override string LocalName { get { return name; } }
        
      public override string Value { get { return value; } }
    }

    internal class RootContext: ContainerContext
    {
      // Cached contexts
      internal TextContext textCache;
      internal WhitespaceContext whitespaceCache;
      internal SignificantWhitespaceContext significantWhitespaceCache;
      internal CommentContext commentCache;
      internal ProcessingInstructionContext processingInstructionCache;

      internal XmlReader reader; // this is a reader that feeds our navigator
      internal XmlNameTable nameTable; // name table
      internal string xmlnsUri; // xmlns namespace uri atom
      internal XmlSpace space = XmlSpace.Default; // whether to preserve spaces
      internal StringBuilder builder; // string builder for calculating value of element
      private int currentDocumentIndex;

      public RootContext(XmlReader reader)
      {
        this.nodeType = XPathNodeType.Root;
        this.root = this;
        this.reader = reader;
        this.nameTable = reader.NameTable;
        this.xmlnsUri = nameTable.Add(NsXmlns);
        this.baseURI = reader.BaseURI;

        // read first time
        while(Read())
        {
          switch(reader.NodeType)
          {
            case XmlNodeType.EndElement:
            case XmlNodeType.EndEntity:
            case XmlNodeType.None:
              return;
            case XmlNodeType.XmlDeclaration:
              continue;
            default:
              firstChild = unknownContext;
              attributeCache = new AttributeContext(this);
              namespaceCache = new NamespaceContext(this);

              return;
          }
        }
      }

      public override Context NextSibling { get { return null; } }

      internal bool Read()
      {
        return reader.Read() && ((reader.NodeType != XmlNodeType.Whitespace) || 
          (space == XmlSpace.Preserve) || reader.Read());
      }

      // Creates context for the current reader's position
      internal Context CreateContext(ContainerContext parent)
      {
        while(true)
        {
          switch(reader.NodeType)
          {
            case XmlNodeType.Element:
            {
              ElementContext element = parent.elementCache;

              if (element != null)
                element.firstChild = null;
              else
                parent.elementCache = element = new ElementContext(this, parent);
              element.Init(++currentDocumentIndex);

              bool isEmpty = reader.IsEmptyElement;

              if (!reader.MoveToFirstAttribute())
              {
                element.firstAttribute = null;
                element.firstNamespace = parent.firstNamespace;
              }
              else
                CreateAttributes(element);
              if (!isEmpty)
              {
                Read();
                if (reader.NodeType != XmlNodeType.EndElement)
                  element.firstChild = unknownContext;
              }
              parent.firstChild = element;
        
              return element;
            }
            case XmlNodeType.EntityReference:
              reader.ResolveEntity();
              break;
            case XmlNodeType.EndEntity:
              break;
            case XmlNodeType.Text:
            case XmlNodeType.CDATA:
              if (textCache == null)
                textCache = new TextContext(this);
              textCache.Init(parent, ++currentDocumentIndex);
              Read();
        
              return textCache;
            case XmlNodeType.XmlDeclaration:
              break;
            case XmlNodeType.ProcessingInstruction:
              if (processingInstructionCache == null)
                processingInstructionCache = new ProcessingInstructionContext(this);
              processingInstructionCache.Init(parent, ++currentDocumentIndex);
              Read();

              return processingInstructionCache;
            case XmlNodeType.Comment:
              if (commentCache == null)
                commentCache = new CommentContext(this);
              commentCache.Init(parent, ++currentDocumentIndex);
              Read();

              return commentCache;
            case XmlNodeType.Whitespace:
              if (whitespaceCache == null)
                whitespaceCache = new WhitespaceContext(this);
              whitespaceCache.Init(parent, ++currentDocumentIndex);
              Read();

              return whitespaceCache;
            case XmlNodeType.SignificantWhitespace:
              if (significantWhitespaceCache == null)
                significantWhitespaceCache = new SignificantWhitespaceContext(this);
              significantWhitespaceCache.Init(parent, ++currentDocumentIndex);
              Read();

              return significantWhitespaceCache;
            default:
              parent.firstChild = null;

              return null;
          }
          Read();
        }
      }
    
      private void CreateAttributes(ElementContext element)
      {
        AttributeContext firstAttribute = null;
        NamespaceContext firstNamespace = null;
        AttributeContext attributeCache = element.attributeCache;
        NamespaceContext namespaceCache = element.namespaceCache;

        do
        {
          if ((object)reader.NamespaceURI == (object)xmlnsUri) 
          {
            NamespaceContext next = namespaceCache;

            namespaceCache = next.namespaceCache;
            if (namespaceCache == null)
              next.namespaceCache = namespaceCache = new NamespaceContext(root);
            next.Init(element, ++currentDocumentIndex);
            next.next = firstNamespace;
            firstNamespace = next;
          }
          else 
          {
            AttributeContext next = attributeCache;

            attributeCache = next.attributeCache;
            if (attributeCache == null)
              next.attributeCache = attributeCache = new AttributeContext(this);
            next.Init(element, ++currentDocumentIndex);
            next.next = firstAttribute;
            firstAttribute = next;
          }
        }
        while(reader.MoveToNextAttribute());

        element.firstAttribute = firstAttribute;

        if (firstNamespace == null)
        {
          element.firstNamespace = element.parent.firstNamespace;
        }
        else
        {
          MergeNamespaces(firstNamespace, element.parent.firstNamespace, namespaceCache);
          element.firstNamespace = firstNamespace;
        }
      }

      private void MergeNamespaces(NamespaceContext firstNamespace,
        NamespaceContext parentNamespace, NamespaceContext namespaceCache)
      {
        // Merge namespaces
        // Find the topmost unconflicting node in old list
        // and remove duplicates from the new list
        NamespaceContext lastNamespace = firstNamespace;
        NamespaceContext unconflictOld = parentNamespace;

        for(NamespaceContext ns = firstNamespace; ns != null; ns = ns.next)
        {
          NamespaceContext confOld = FindByName(unconflictOld, ns.localName);

          if (confOld != null) 
          {
            // Even if confOld.Value == ns.Value we have to replace this ns context
            // because new has different docIndex and be sorted differentely
            unconflictOld = confOld.next;
          }
          lastNamespace = ns;
        }

        // now mirgedList has only unconfliced part of parent namespace list
        NamespaceContext mirgedList = unconflictOld;

        // from interval (unconflictOld, parentNamespace] attached to mirgedList clones 
        // of all unconflicting context
        for (NamespaceContext ns = parentNamespace; ns != unconflictOld; ns = ns.next) 
        {
          if (FindByName(firstNamespace, ns.Name) == null) 
          {
            NamespaceContext clone = namespaceCache;

            namespaceCache = clone.namespaceCache;
            if (namespaceCache == null)
              clone.namespaceCache = namespaceCache = new NamespaceContext(this);
            clone.Init(ns);
            clone.next = mirgedList;
            mirgedList = clone;
          }
          else 
          {
            // Just ignore it
          }
        }

        // now we can attach new nodes
        lastNamespace.next = mirgedList;
      }

      private static NamespaceContext FindByName(NamespaceContext list, string name) 
      {
        while(list != null && list.localName != name) 
        {
          list = list.next;
        }

        return list;
      }
    }

    private Context context; // determines position of the navigator
    private int documentIndex; // determines document index of the context

    /// <summary>
    ///   Initializes a new instance of the ForwardXPathNavigator class.
    /// </summary>
    /// <param name = "reader">
    ///   The XmlReader to read from while navigating.
    /// </param>
		public ForwardXPathNavigator(XmlReader reader)
		{
      if (reader == null)
        throw new ArgumentException(readerIsNotSpecified);

      context = new RootContext(reader);
    }

    /// <summary>
    ///   Initializes a new instance of the ForwardXPathNavigator class.
    /// </summary>
    /// <param name = "reader">
    ///   The XmlReader to read from while navigating.
    /// </param>
    /// <param name = "space">
    ///   An XmlSpace value indicating whether to preserve whitespace. 
    ///   Setting this to XmlSpace.Default preserves only significant white space; 
    ///   XmlSpace.Preserve preserves all white space.
    /// </param>
    public ForwardXPathNavigator(XmlReader reader, XmlSpace space): this(reader)
    {
      context.root.space = space;
    }

    /// <summary>
    /// Copy constructor of the ForwardXPathNavigator class.
    /// </summary>
    /// <param name="other">navigator to use for copy.</param>
    public ForwardXPathNavigator(ForwardXPathNavigator other)
    {
      this.context = other.context;
      this.documentIndex = other.documentIndex;
    }

    /// <summary>
    ///   Gets the XmlReader used to construct this ForwardXPathNavigator.
    /// </summary>
    /// <value>The XmlReader specified in the constructor.</value>
    /// <remarks>
    ///   This property allows the user to access properties specific to 
    ///   the supplied reader. It returns a typed XmlReader. Changing the 
    ///   state of the returned reader (for example, calling one of the 
    ///   read methods) can lead to unpredictable results.
    /// </remarks>
    public XmlReader Reader { get { return context.root.reader; } }

    /// <summary>
    ///   Defines how to handle spaces.
    /// </summary>
    /// <value>
    ///   An XmlSpace value indicating whether to preserve whitespace. 
    ///   Setting this to XmlSpace.Default preserves only significant white space; 
    ///   XmlSpace.Preserve preserves all white space.
    ///   By default Space is XmlSpace.Default.
    /// </value>
    public XmlSpace Space
    {
      get { return context.root.space; }
      set { context.root.space = value; }
    }

    // XPathNavigator methods
    /// <summary>
    ///   Returns clone of the navigator.
    /// </summary>
    /// <returns>
    ///   Returns clone of the navigator.
    /// </returns>
    public override XPathNavigator Clone()
    {
      return new ForwardXPathNavigator(this);
    }

    /// <summary>Gets the type of the current node.</summary>
    /// <value>One of the XPathNodeType values.</value>
    /// <remarks><seealso cref = "XPathNavigator.NodeType"/></remarks>
    public override XPathNodeType NodeType
    { 
      get 
      { 
        return context.nodeType;
      } 
    }

    /// <summary>
    ///   Gets the name of the current node without the namespace prefix.
    /// </summary>
    /// <value>
    ///   The local name of the current node. The name returned depends on 
    ///   the NodeType of the current node. 
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.LocalName"/></remarks>
    public override string LocalName
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.LocalName;
      } 
    }

    /// <summary>
    ///   Gets the namespace URI (as defined in the W3C Namespace Specification) 
    ///   of the current node.
    /// </summary>
    /// <value>
    ///   The namespace URI of the current node or an empty string if the current 
    ///   node does not have one.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.NamespaceURI"/></remarks>
    public override string NamespaceURI
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.NamespaceURI;
      } 
    }

    /// <summary>
    ///   Gets the qualified name of the current node.
    /// </summary>
    /// <value>
    ///   The qualified name of the current node. The name returned depends 
    ///   on the NodeType of the current node.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.Name"/></remarks>
    public override string Name
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.Name;
      } 
    }

    /// <summary>
    ///   Gets the prefix associated with the current node.
    /// </summary>
    /// <value>The prefix associated with the current node.</value>
    /// <remarks><seealso cref = "XPathNavigator.Prefix"/></remarks>
    public override  string Prefix
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.Prefix;
      } 
    }

    /// <include file = 'doc\XPathNavigator.uex' path = 'docs/doc[@for = "XPathNavigator.Value"]/*' />
    public override string Value
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.Value;
      } 
    }

    /// <summary>
    /// Gets the base URI for the current node.
    /// </summary>
    /// <value>The base URI for the current node or empty string if there is no value.</value>
    /// <remarks><seealso cref = "XPathNavigator.Prefix"/></remarks>
    public override String BaseURI
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.BaseUri;
      } 
    }

    /// <summary>
    ///   Gets the xml:lang scope for the current node.
    /// </summary>
    /// <value>
    ///   The value of the xml:lang scope or empty string if the current 
    ///   node has no value to return.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.XmlLang"/></remarks>
    public override String XmlLang
    {
      get 
      { 
        Context current = context;

        if (documentIndex != current.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        do 
        {
          AttributeContext attr = current.GetAttribute(Lang, NsXml);
        
          if (attr != null) return attr.Value; 

          current = current.parent;
        } 
        while(current != null);

        return string.Empty;
      }
    }
        
    /// <summary>
    ///   Gets a value indicating whether the current node is an empty element 
    ///   (for example, &lt;MyElement/>).
    /// </summary>
    /// <value>
    ///   true if the current node is an element (NodeType equals XPathNodeType.Element) 
    ///   that ends with "/>"; otherwise, false.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.IsEmptyElement"/></remarks>
    public override bool IsEmptyElement { get { return HasChildren; } }

    /// <summary>
    ///   Gets the XmlNameTable associated with this implementation.
    /// </summary>
    /// <value>
    ///   An XmlNameTable.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.NameTable"/></remarks>
    public override XmlNameTable NameTable { get { return context.root.nameTable; } }

    // Attributes
    /// <summary>
    ///   Gets a value indicating whether the element node has any attributes.
    /// </summary>
    /// <value>
    ///   true if the element node has attributes; 
    ///   false if the node has no attributes or if the XPathNavigator 
    ///   is not positioned on an element node.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.HasAttributes"/></remarks>
    public override bool HasAttributes
    { 
      get 
      { 
        if (documentIndex != context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        return context.HasAttributes;
      } 
    }

    /// <summary>
    ///   Gets the value of the attribute with the specified LocalName and NamespaceURI.
    /// </summary>
    /// <param name = "localName">The local name of the attribute.</param>
    /// <param name = "NamespaceURI">The namespace URI of the attribute.</param>
    /// <returns>
    ///   The value of the specified attribute; 
    ///   String.Empty if a matching attribute is not found or 
    ///   if the navigator is not positioned on an element node.
    /// </returns>
    /// <remarks><seealso cref = "XPathNavigator.GetAttribute"/></remarks>
    public override string GetAttribute(string localName, string namespaceURI)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      AttributeContext attr = context.GetAttribute(localName, namespaceURI);

      return attr != null ? attr.value : string.Empty;
    }

    /// <summary>
    ///   Moves to the attribute with matching localName and namespaceURI.
    /// </summary>
    /// <param name = "localName">The local name of the attribute.</param>
    /// <param name = "namespaceURI">The namespace URI of the attribute.</param>
    /// <returns>
    ///   true if the attribute is found; otherwise, false. 
    ///   If false, the position of the navigator does not change.
    /// </returns>
    /// <remarks><seealso cref = "XPathNavigator.MoveToAttribute"/></remarks>
    public override bool MoveToAttribute(string localName, string namespaceURI)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      Context attr = context.GetAttribute(localName, namespaceURI);

      if (attr != null)
      {
        context = attr;
        documentIndex = attr.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>
    ///   Moves to the first attribute.
    /// </summary>
    /// <returns>
    ///   true if the navigator is successful moving to the first attribute; otherwise, false.
    /// </returns>
    /// <remarks>
    ///   If the navigator is not currently positioned on an element, this method returns 
    ///   false and the position of the navigator does not change.
    ///   <seealso cref = "XPathNavigator.MoveToFirstAttribute"/>
    /// </remarks>
    public override bool MoveToFirstAttribute()
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      Context attr = context.FirstAttribute;

      if (attr != null)
      {
        context = attr;
        documentIndex = attr.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>
    ///   Moves to the next attribute.
    /// </summary>
    /// <returns>
    ///   true if the navigator is successful moving to the next attribute; 
    ///   false if there are no more attributes.
    /// </returns>
    /// <remarks>
    ///   If the navigator is not currently positioned on an attribute, this method returns 
    ///   false and the position of the navigator does not change.
    ///   <seealso cref = "XPathNavigator.MoveToNextAttribute"/>
    /// </remarks>
    public override bool MoveToNextAttribute()
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      Context attr = context.NextAttribute;

      if (attr != null)
      {
        context = attr;
        documentIndex = attr.documentIndex;

        return true;
      }

      return false;
    }

    // Namespaces
    /// <summary>
    ///   Returns the value of the namespace node corresponding to the specified local name.
    /// </summary>
    /// <param name = "name">The local name of the namespace node.</param>
    /// <returns>
    ///   The value of the namespace node; String.Empty if a matching namespace 
    ///   node is not found or if the navigator is not positioned on an element node.
    ///   </returns>
    /// <remarks>
    ///   <seealso cref = "XPathNavigator.GetNamespace"/>
    /// </remarks>
    public override string GetNamespace(string name)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      NamespaceContext ns = context.GetNamespace(name);
    
      if (ns != null) return ns.value;
      if (name == "xml") return NsXml;
      if (name == "xmlns") return NsXmlns;
      
      return string.Empty;			
    }

    /// <summary>
    ///   This method always return false due to XmlReader does not public
    ///   NamespaceManager
    /// </summary>
    public override bool MoveToNamespace(string name)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      NamespaceContext ns = context.GetNamespace(name);

      if (ns != null)
      {
        if (ns.parent != context)
        {
          ContainerContext parentContext = (ContainerContext)context;
          NamespaceContext copy = parentContext.namespaceCache;

          if (copy.namespaceCache == null)
            copy.namespaceCache = new NamespaceContext(context.root);

          copy.Init(ns);
          copy.parent = parentContext;
          ns = copy;
        }

        context = ns;
        documentIndex = ns.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>
    ///   This method always return false due to XmlReader does not public
    ///   NamespaceManager
    /// </summary>
    public override bool MoveToFirstNamespace(XPathNamespaceScope namespaceScope)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      NamespaceContext ns = context.FirstNamespace;

      if (ns != null)
      {
        Context parent = ns.parent;

        if (parent == context || namespaceScope != XPathNamespaceScope.Local)
        {
          if (parent != context)
          {
            ContainerContext parentContext = (ContainerContext)context;
            NamespaceContext copy = parentContext.namespaceCache;

            if (copy.namespaceCache == null)
              copy.namespaceCache = new NamespaceContext(context.root);

            copy.Init(ns);
            copy.parent = parentContext;
            ns = copy;
          }
          
          context = ns;
          documentIndex = ns.documentIndex;

          return true;
        }
      }
      
      return false;
    }

    /// <summary>
    ///   This method always return false due to XmlReader does not public
    ///   NamespaceManager
    /// </summary>
    public override bool MoveToNextNamespace(XPathNamespaceScope namespaceScope)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      NamespaceContext ns = context.NextNamespace;

      if (ns != null)
      {
        ContainerContext parent = ns.parent;

        // We do not suppose that namespaceScope is different from the corresponding value
        // in call to MoveToFirstNamespace.
        if (parent == context.parent || namespaceScope != XPathNamespaceScope.Local)
        {
          if (parent != context.parent)
          {
            ContainerContext parentContext = context.parent;
            NamespaceContext copy = parentContext.namespaceCache;

            if (copy.namespaceCache == null)
              copy.namespaceCache = new NamespaceContext(context.root);

            copy.Init(ns);
            copy.parent = parentContext;
            ns = copy;
          }

          context = ns;
          documentIndex = ns.documentIndex;

          return true;
        }
      }
      
      return false;
    }

    // Tree
    /// <summary>
    ///   Moves to the next sibling of the current node.
    /// </summary>
    /// <returns>
    ///   true if the navigator is successful moving to the next sibling node; 
    ///   false if there are no more siblings or if the navigator is currently 
    ///   positioned on an attribute node. If false, the position of the navigator 
    ///   is unchanged.
    /// </returns>
    /// <remarks>
    ///   <seealso cref = "XPathNavigator.MoveToNext"/>
    /// </remarks>
    public override bool MoveToNext()
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      Context current = context.NextSibling;

      if (current != null)
      {
        context = current;
        documentIndex = current.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>
    ///   Moves to the previous sibling of the current node.
    /// </summary>
    /// <returns>
    ///   true if the navigator is successful moving to the previous sibling node; 
    ///   false if there is no previous sibling or if the navigator is currently 
    ///   positioned on an attribute node.
    /// </returns>
    /// <remarks>
    ///   Implementation of the method throws exception whenever this is not
    ///   first node due to previous data never is available
    /// </remarks>
    public override bool MoveToPrevious() 
    { 
      throw new InvalidOperationException(operationIsNotAvailable);
    }

    /// <summary>
    ///   Moves to the first sibling of the current node.
    /// </summary>
    /// <returns>
    ///   true if the navigator is successful moving to the first sibling node; 
    ///   false if there is no first sibling or if the navigator is currently 
    ///   positioned on an attribute node.
    /// </returns>
    /// <remarks>
    ///   Implementation of the method throws exception whenever this is not
    ///   first node due to previous data never is available
    /// </remarks>
    public override bool MoveToFirst()    
    { 
      throw new InvalidOperationException(operationIsNotAvailable);
    }

    /// <summary>
    ///   Gets a value indicating whether the current node has child nodes.
    /// </summary>
    /// <value>
    ///   true if the current node has child nodes; otherwise, false. 
    ///   If the navigator is positioned on a node which cannot have children, 
    ///   this property always returns false.
    /// </value>
    /// <remarks><seealso cref = "XPathNavigator.HasChildren"/></remarks>
    public override bool HasChildren
    { 
      get 
      { 
        // We can remove this check as we have the same one in MoveToFirstChild() and
        // in ContextIterator.
        // if (documentIndex != context.documentIndex)
        //   throw new InvalidOperationException(operationIsNotAvailable);

        return context.firstChild != null; 
      } 
    }

    /// <summary>
    ///   Moves to the first child of the current node.
    /// </summary>
    /// <returns>true if there is a first child node; otherwise false.</returns>
    /// <remarks>
    ///   <seealso cref = "XPathNavigator.HasChildren"/>
    /// </remarks>
    public override bool MoveToFirstChild()
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      Context current = context.FirstChild;

      if (current != null)
      {
        context = current;
        documentIndex = current.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>
    ///   Moves to the parent of the current node.
    /// </summary>
    /// <returns>true if there is a parent node; otherwise false.</returns>
    /// <remarks><seealso cref = "XPathNavigator.MoveToParent"/></remarks>
    public override bool MoveToParent()
    {
      Context parent = context.parent;

      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      if (parent != null)
      {
        context = parent;
        documentIndex = parent.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>Moves to the root node to which the current node belongs.</summary>
    /// <remarks><seealso cref = "XPathNavigator.MoveToRoot"/></remarks>
    public override void MoveToRoot()
    {
      context = context.root;
      documentIndex = 0;
    }

    /// <summary>
    ///   Moves to the same position as the specified XPathNavigator.
    /// </summary>
    /// <param name = "other">
    ///   The XPathNavigator positioned on the node that you want to move to. 
    /// </param>
    /// <returns>
    ///   true if successful; otherwise false. 
    ///   If false, the position of the navigator is unchanged.
    /// </returns>
    /// <remarks><seealso cref = "XPathNavigator.MoveTo"/></remarks>
    public override bool MoveTo(XPathNavigator other)
    {
      ForwardXPathNavigator navigator = other as ForwardXPathNavigator;

      if (navigator != null)
      {
        context = navigator.context;
        documentIndex = navigator.documentIndex;

        return true;
      }
      
      return false;
    }

    /// <summary>MoveToId is not supported</summary>
    public override bool MoveToId(string id) { return false; }

    /// <summary>
    ///   Determines whether the current XPathNavigator is at the same position 
    ///   as the specified XPathNavigator.
    /// </summary>
    /// <param name = "other">The XPathNavigator that you want to compare against.</param>
    /// <returns>
    ///   true if the two navigators have the same position; otherwise, false.
    /// </returns>
    /// <remarks><seealso cref = "XPathNavigator.IsSamePosition"/></remarks>
    public override bool IsSamePosition(XPathNavigator other)
    {
      ForwardXPathNavigator navigator = other as ForwardXPathNavigator;

      return navigator != null && context.root == navigator.context.root &&
        documentIndex == navigator.documentIndex;
    }

    /// <summary>
    ///   Compares the position of the current navigator with the position of the specified 
    ///   XPathNavigator.
    /// </summary>
    /// <param name = "nav">The XPathNavigator to compare against.</param>
    /// <returns>
    ///   An XmlNodeOrder value representing the comparative position of 
    ///   the two navigators. The following table describes the XmlNodeOrder enumeration.
    /// </returns>
    /// <remarks><seealso cref = "XPathNavigator.ComparePosition"/></remarks>
    public override XmlNodeOrder ComparePosition(XPathNavigator other) 
    {
      ForwardXPathNavigator navigator = other as ForwardXPathNavigator;

      if (navigator != null && context.root == navigator.context.root)
      {
        int delta = documentIndex - navigator.documentIndex;

        if (delta == 0)
          return XmlNodeOrder.Same;
        if (delta < 0)
          return XmlNodeOrder.Before;

        return XmlNodeOrder.After;
      }

      return XmlNodeOrder.Unknown;
    }

    /// <summary>
    ///   Determines whether the specified XPathNavigator is a descendant of the current 
    ///   XPathNavigator.
    /// </summary>
    /// <param name = "nav">The XPathNavigator that you want to compare against.</param>
    /// <returns>
    ///   true if nav is a descendant of the current navigator; otherwise false.
    /// </returns>
    /// <remarks><seealso cref = "XPathNavigator.IsDescendant"/></remarks>
    public override bool IsDescendant(XPathNavigator other)
    {
      if (documentIndex != context.documentIndex)
        throw new InvalidOperationException(operationIsNotAvailable);

      ForwardXPathNavigator navigator = other as ForwardXPathNavigator;

      if (navigator != null && context.root == navigator.context.root)
      {
        if (navigator.documentIndex != navigator.context.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        for(Context current = context.parent; current != null; current = current.parent)
        {
          if (current == navigator.context) return true;
        }
      }

      return false;
    }

    /// <summary>
    ///   Selects all the child element nodes of the current node with the supplied 
    ///   local name and namespace URI.
    /// </summary>
    /// <param name = "name">The local name of the child nodes.</param>
    /// <param name = "namespaceURI">The namespace URI of the child nodes.</param>
    /// <returns>An XPathNodeIterator pointing to the selected nodes.</returns>
    /// <remarks><seealso cref = "XPathNavigator.SelectChildren"/></remarks>
    public override XPathNodeIterator SelectChildren(string name, string namespaceURI) 
    {
      return new ContextIterator(this, name, namespaceURI);
    }

    public override XPathNodeIterator SelectChildren(XPathNodeType nodeType) 
    {
      return new ContextIterator(this, nodeType);
    }

    internal class ContextIterator: XPathNodeIterator
    {
      private bool next;
      private ForwardXPathNavigator navigator;
      private int position;
      private string name;
      private string namespaceURI;
      private XPathNodeType nodeType;

      public ContextIterator(ForwardXPathNavigator navigator,
        string name, string namespaceURI)
      { 
        this.navigator = new ForwardXPathNavigator(navigator);
        this.nodeType = XPathNodeType.Element;

        XmlNameTable nameTable = navigator.context.root.nameTable;

        this.name = nameTable.Add(name);
        this.namespaceURI = nameTable.Add(namespaceURI);
      }

      public ContextIterator(ForwardXPathNavigator navigator, XPathNodeType nodeType)
      { 
        this.navigator = new ForwardXPathNavigator(navigator);
        this.nodeType = nodeType;
      }

      public override XPathNavigator Current { get { return navigator; } }

      public override int CurrentPosition { get { return position; } }

      public override XPathNodeIterator Clone() { return this; }

      public override bool MoveNext() 
      {
        Context current = navigator.context;

        if (navigator.documentIndex != current.documentIndex)
          throw new InvalidOperationException(operationIsNotAvailable);

        bool next = this.next;

        if (next)
        {
          current = navigator.context.NextSibling;

          if (current == null) return false;
        }
        else
        {
          current = navigator.context.FirstChild;

          if (current == null) return false;
          this.next = true;
        }

        do
        {
          if (nodeType == XPathNodeType.All)
            goto found;
          else if ((object)name != null)
          {
            if (current.IsElement(name, namespaceURI))
              goto found;
            else
              if (next) return false;
          }
          else if (nodeType == current.nodeType)
            goto found;

          current = current.NextSibling;
        }
        while(current != null);

        return false;

      found:
        position++;
        navigator.context = current;
        navigator.documentIndex = current.documentIndex;
          
        return true;
      }
    }
  }
}
