﻿namespace NesterovskyBros.DataStream
{
  using System;
  using System.IO;
  using System.Threading;
  using System.Threading.Tasks;
  using System.Diagnostics;

  /// <summary>
  /// A class that allows to access steam reasource multiple times,
  /// while original stream is read once only.
  /// </summary>
  /// <remarks>
  /// The class is safe to use in multiple threads.
  /// </remarks>
  public class StreamResource: IDisposable
  {
    /// <summary>
    /// Settings for the StreamResource.
    /// </summary>
    public class Settings
    {
      public Settings()
      {
        BufferSize = 8192;
        CreateTempFileTryCount = 2;
        PoolTimeout = 100;
      }

      /// <summary>
      /// A buffer size.
      /// Default value is 8192 bytes.
      /// </summary>
      public int BufferSize { get; set; }

      /// <summary>
      /// Number of tries to create a temp cache file.
      /// Default value is 3.
      /// </summary>
      public int CreateTempFileTryCount { get; set; }

      /// <summary>
      /// A folder to store cache data. 
      /// If value is not specified then a temp folder is used.
      /// Default value is null.
      /// </summary>
      public string DataFolder { get; set; }

      /// <summary>
      /// A maximum number of milliseconds to allow between data consumption time
      /// and data read time.
      /// If data is read faster than it's consumed, and the difference exceeds 
      /// PoolTimeout value then the data is being started to cache in 
      /// a greedy way.
      /// Default value is 100 milliseconds.
      /// </summary>
      public int PoolTimeout { get; set; }
    }

    /// <summary>
    /// Creates a StreamSource instance.
    /// </summary>
    /// <param name="source">
    /// A function that returns source as an input stream.
    /// </param>
    /// <param name="settings">Optional settings.</param>
    public StreamResource(Func<Stream> source, Settings settings = null)
    {
      if (source == null)
      {
        throw new ArgumentException("source");
      }

      this.settings = settings ?? new Settings();

      Run(
        output =>
        {
          using(var input = source())
          {
            input.CopyTo(output, this.settings.BufferSize);
          }
        });
    }

    /// <summary>
    /// Creates a StreamSource instance.
    /// </summary>
    /// <param name="source">
    /// A function that writes source data into an output stream.
    /// </param>
    /// <param name="settings">Optional settings.</param>
    public StreamResource(Action<Stream> source,  Settings settings = null)
    {
      if (source == null)
      {
        throw new ArgumentException("source");
      }

      this.settings = settings ?? new Settings();

      Run(source);
    }

    /// <summary>
    /// Disposes StreamResource instance.
    /// </summary>
    public void Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    /// <summary>
    /// Gets the number of currently opened streams over this StreamResource.
    /// </summary>
    public int UseCount
    {
      get { return useCount; }
    }

    /// <summary>
    /// Gets an input stream.
    /// </summary>
    /// <remarks>
    /// Stream can be read and positioned only.
    /// </remarks>
    /// <param name="shared">
    /// Indicates that this StreamResouce should be disposed when returned
    /// stream is closed and there are no more currently opened cache streams.
    /// </param>
    /// <returns>A input stream.</returns>
    public Stream GetStream(bool shared = false)
    {
      var succeed = false;

      try
      {
        var result = new CacheStream(this, shared);

        succeed = true;

        return result;
      }
      finally
      {
        if (!succeed)
        {
          Dispose();
        }
      }
    }

    /// <summary>
    /// Disposes StreamResource instance.
    /// </summary>
    protected virtual void Dispose(bool disposing)
    {
      if (disposing)
      {
        if (Interlocked.CompareExchange(ref disposed, 1, 0) != 0)
        {
          return;
        }

        try
        {
          Eof();
        }
        finally
        {
          try
          {
            var stream = cache;

            if (stream != null)
            {
              stream.Dispose();
            }
          }
          finally
          {
            inputEvent.Dispose();
          }
        }
      }
    }

    /// <summary>
    /// Starts caching an input.
    /// </summary>
    /// <param name="source">An output source.</param>
    private void Run(Action<Stream> source)
    {
      Task.Factory.StartNew(
        () =>
        {
          using(var output = new SourceStream(this))
          {
            readWatch.Start();
            source(output);
          }
        },
        TaskCreationOptions.LongRunning);
    }

    /// <summary>
    /// A task that competes when cache stream is flushed.
    /// </summary>
    private Task GetFlushTask()
    {
      CheckNotDisposed();

      var completion = flushCompletion;

      if (completion == null)
      {
        var newCompletion = new TaskCompletionSource<bool>();

        completion = Interlocked.CompareExchange(
          ref flushCompletion,
          newCompletion,
          null) ?? newCompletion;

        if ((disposed != 0) || eof)
        {
          newCompletion.TrySetResult(false);
        }
      }

      return completion.Task;
    }

    /// <summary>
    /// Creates a cache stream.
    /// </summary>
    private FileStream GetCacheStream()
    {
      var stream = cache;

      if (stream != null)
      {
        return stream;
      }

      Exception ex = null;
      var folder = settings.DataFolder ?? Path.GetTempPath();
      var count = settings.CreateTempFileTryCount;

      if (count <= 0)
      {
        count = 1;
      }

      for(var i = 0; i < count; ++i)
      {
        CheckNotDisposed();

        var path = Path.Combine(folder, Path.GetRandomFileName());
        var newStream = new FileStream(
          path,
          FileMode.CreateNew,
          FileAccess.Write,
          FileShare.Read,
          1,
          FileOptions.DeleteOnClose);

        try
        {
          stream = 
            Interlocked.CompareExchange(ref cache, newStream, null) ??
            newStream;

          CheckNotDisposed();

          if (newStream == stream)
          {
            newStream = null;
          }

          return stream;
        }
        catch(IOException e)
        {
          ex = e;
        }
        finally
        {
          if (newStream != null)
          {
            newStream.Dispose();
          }
        }
      }

      throw ex;
    }

    /// <summary>
    /// Marks end of input data.
    /// </summary>
    private void Eof()
    {
      eof = true;

      var completion = flushCompletion;

      if (completion != null)
      {
        completion.TrySetResult(false);
      }

      readWatch.Stop();
      consumeWatch.Stop();
    }

    /// <summary>
    /// Writes block of data into the cache.
    /// </summary>
    /// <param name="buffer">A data to cache</param>
    /// <param name="offset">Data offset in the buffer.</param>
    /// <param name="count">Data size.</param>
    /// <param name="eof">End indicator, if v</param>
    private void Write(byte[] buffer, int offset, int count)
    {
      readWatch.Stop();
      consumeWatch.Start();

      var stream = GetCacheStream();

      stream.Write(buffer, offset, count);
      stream.Flush();

      Interlocked.Add(ref position, count);

      var completion = Interlocked.Exchange(ref flushCompletion, null);

      if (completion != null)
      {
        completion.TrySetResult(true);
      }

      var delta = 
        consumeWatch.ElapsedMilliseconds - readWatch.ElapsedMilliseconds;
      var timeout = settings.PoolTimeout;

      if ((delta > 0) && (delta < timeout))
      {
        inputEvent.Wait((int)(timeout - delta));
        inputEvent.Reset();
      }

      consumeWatch.Stop();
      readWatch.Start();
    }

    /// <summary>
    /// Checks that StreamResource instance is not disposed.
    /// </summary>
    private void CheckNotDisposed()
    {
      if (disposed != 0)
      {
        throw new ObjectDisposedException("StreamResource");
      }
    }

    /// <summary>
    /// A cache stream.
    /// </summary>
    private class CacheStream: Stream
    {
      /// <summary>
      /// Creates an instance of a cache stream.
      /// </summary>
      /// <param name="resource"><see cref="StreamResource"/> instance.</param>
      /// <param name="shared">
      /// Indicates that StreamResouce should be disposed when this stream 
      /// is closed and there are no more currently opened cache streams.
      /// </param>
      public CacheStream(StreamResource resource, bool shared)
      {
        this.resource = resource;
        this.shared = shared;
        Interlocked.Increment(ref resource.useCount);
      }

      /// <summary>
      /// Returns true, as stream can be read.
      /// </summary>
      public override bool CanRead
      {
        get { return true; }
      }

      /// <summary>
      /// Returns true.
      /// </summary>
      public override bool CanSeek
      {
        get { return true; }
      }

      /// <summary>
      /// Returns false, as stream does not support write.
      /// </summary>
      public override bool CanWrite
      {
        get { return false; }
      }

      /// <summary>
      /// Flush does nothing.
      /// </summary>
      public override void Flush()
      {
      }

      /// <summary>
      /// No length is known in advance.
      /// </summary>
      public override long Length
      {
        get 
        {
          while(resource.eof)
          {
            resource.GetFlushTask().Wait();
          }

          return resource.position;
        }
      }

      /// <summary>
      /// Positions the stream.
      /// </summary>
      public override long Position
      {
        get { return GetCacheStream().Position; }
        set { GetCacheStream().Position = value; }
      }

      /// <summary>
      /// Begins an asynchronous read operation.
      /// </summary>
      /// <param name="buffer">
      /// The buffer to read the data into.
      /// </param>
      /// <param name="offset">
      /// The byte offset in buffer at which to begin writing data read from the stream.
      /// </param>
      /// <param name="count">
      /// The maximum number of bytes to read.
      /// </param>
      /// <param name="callback">
      /// An optional asynchronous callback, to be called when 
      /// the read is complete.
      /// </param>
      /// <param name="state">
      /// A user-provided object that distinguishes this particular 
      /// asynchronous read request from other requests.
      /// </param>
      /// <returns>
      /// An System.IAsyncResult that represents the asynchronous read, 
      /// which could still be pending.
      /// </returns>
      public override IAsyncResult BeginRead(
        byte[] buffer, 
        int offset, 
        int count, 
        AsyncCallback callback, 
        object state)
      {
        var completionSource = new TaskCompletionSource<int>(state);
        Action read = null;

        read = () =>
        {
          try
          {
            var stream = GetCacheStream();

            stream.BeginRead(
              buffer,
              offset,
              count,
              ar =>
              {
                try
                {
                  var readCount = stream.EndRead(ar);

                  if ((readCount == 0) && !resource.eof)
                  {
                    resource.inputEvent.Set();

                    resource.GetFlushTask().ContinueWith(
                      t =>
                      {
                        if (t.IsFaulted)
                        {
                          completionSource.TrySetException(t.Exception);

                          return;
                        }

                        read();
                      });
                  }
                  else
                  {
                    completionSource.TrySetResult(readCount);

                    if (callback != null)
                    {
                      callback(completionSource.Task);
                    }
                  }
                }
                catch(Exception e)
                {
                  completionSource.TrySetException(e);
                }
              },
              null);
          }
          catch (Exception e)
          {
            completionSource.TrySetException(e);

            return;
          }
        };

        if (cache == null)
        {
          resource.GetFlushTask().ContinueWith(
            t =>
            {
              if (t.IsFaulted)
              {
                completionSource.TrySetException(t.Exception);

                return;
              }

              read();
            });
        }
        else
        { 
          read();
        }
        
        return completionSource.Task;
      }

      /// <summary>
      /// Waits for the pending asynchronous read to complete.
      /// </summary>
      /// <param name="asyncResult">
      /// The reference to the pending asynchronous request to finish.
      /// </param>
      /// <returns>
      /// The number of bytes read from the stream, between zero (0) and 
      /// the number of bytes you requested. Streams return zero (0) only
      /// at the end of the stream, otherwise, they should block until at 
      /// least one byte is available.
      /// </returns>
      public override int EndRead(IAsyncResult asyncResult)
      {
        Task<int> task = (Task<int>)asyncResult;

        return task.Result;
      }

      /// <summary>
      /// Reads a sequence of bytes from the current stream and advances 
      /// the position within the stream by the number of bytes read.
      /// </summary>
      /// <param name="buffer">A buffer to populate with data.</param>
      /// <param name="offset">A start offset in the buffer to store data.</param>
      /// <param name="count">A maximum number of bytes to write.</param>
      /// <returns>The total number of bytes read into the buffer. 
      /// This can be less than the number of bytes requested if that 
      /// many bytes are not currently available, or zero (0) if the end 
      /// of the stream has been reached.
      /// </returns>
      public override int Read(byte[] buffer, int offset, int count)
      {
        return EndRead(BeginRead(buffer, offset, count, null, null));
      }

      /// <summary>
      /// No seek is supported.
      /// </summary>
      public override long Seek(long offset, SeekOrigin origin)
      {
        return GetCacheStream().Seek(offset, origin);
      }

      /// <summary>
      /// No operation is supported.
      /// </summary>
      public override void SetLength(long value)
      {
        throw new NotSupportedException();
      }

      /// <summary>
      /// No write operation is supported.
      /// </summary>
      public override void Write(byte[] buffer, int offset, int count)
      {
        throw new NotSupportedException();
      }

      /// <summary>
      /// Disposes this instance.
      /// </summary>
      /// <param name="disposing">A dispose indicator.</param>
      protected override void Dispose(bool disposing)
      {
        try
        {
          if (disposing)
          {
            if (Interlocked.CompareExchange(ref disposed, 1, 0) != 0)
            {
              return;
            }

            try
            {
              var stream = this.cache;

              if (stream != null)
              {
                stream.Dispose();
              }
            }
            finally
            {
              if ((Interlocked.Decrement(ref resource.useCount) == 0) && 
                shared)
              {
                resource.Dispose();
              }
            }
          }
        }
        finally
        {
          base.Dispose(disposing);
        }
      }

      /// <summary>
      /// Gets a cache stream.
      /// </summary>
      private FileStream GetCacheStream()
      {
        CheckNotDisposed();

        var stream = cache;

        if (stream != null)
        {
          return stream;
        }

        var newStream = new FileStream(
          resource.cache.Name,
          FileMode.Open,
          FileAccess.Read,
          FileShare.ReadWrite | FileShare.Delete,
          resource.settings.BufferSize,
          FileOptions.Asynchronous);

        try
        {
          stream = Interlocked.CompareExchange(ref cache, newStream, null) ??
            newStream;

          CheckNotDisposed();

          if (newStream == stream)
          {
            newStream = null;
          }

          return stream;
        }
        finally
        {
          if (newStream != null)
          {
            newStream.Dispose();
          }
        }
      }

      /// <summary>
      /// Checks that CacheStream instance is not disposed.
      /// </summary>
      private void CheckNotDisposed()
      {
        if (resource.disposed != 0)
        {
          Dispose();
        }

        if (disposed != 0)
        {
          throw new ObjectDisposedException("CacheStream");
        }
      }

      /// <summary>
      /// A disposed indicator.
      /// </summary>
      private int disposed;

      /// <summary>
      /// A cache stream;
      /// </summary>
      private FileStream cache;

      /// <summary>
      /// Indicates that StreamResouce should be disposed when this stream 
      /// is closed and there are no more currently opened cache streams.
      /// </summary>
      private readonly bool shared;

      /// <summary>
      /// A stream resource.
      /// </summary>
      private readonly StreamResource resource;
    }

    /// <summary>
    /// An output stream which is used to write input data.
    /// </summary>
    private class SourceStream: Stream
    { 
      /// <summary>
      /// Creates an instance of a cache stream.
      /// </summary>
      /// <param name="resource"><see cref="StreamResource"/> instance.</param>
      public SourceStream(StreamResource resource)
      {
        this.resource = resource;
      }

      /// <summary>
      /// Returns true, as stream can be read.
      /// </summary>
      public override bool CanRead
      {
        get { return false; }
      }

      /// <summary>
      /// Returns false, as stream does not support seek.
      /// </summary>
      public override bool CanSeek
      {
        get { return false; }
      }

      /// <summary>
      /// Returns false, as stream does not support write.
      /// </summary>
      public override bool CanWrite
      {
        get { return true; }
      }

      /// <summary>
      /// Flush does nothing.
      /// </summary>
      public override void Flush()
      {
      }

      /// <summary>
      /// No length is known in advance.
      /// </summary>
      public override long Length
      {
        get { throw new NotSupportedException(); }
      }

      /// <summary>
      /// No positioning is supported.
      /// </summary>
      public override long Position
      {
        get 
        { 
          return resource.position; 
        }
        set 
        {
          if (resource.position == value)
          {
            return;
          }

          throw new NotSupportedException(); 
        }
      }

      /// <summary>
      /// No read is supported
      /// </returns>
      public override int Read(byte[] buffer, int offset, int count)
      {
        throw new NotSupportedException();
      }

      /// <summary>
      /// No seek is supported.
      /// </summary>
      public override long Seek(long offset, SeekOrigin origin)
      {
        switch(origin)
        {
          case SeekOrigin.Begin:
          {
            if (offset == resource.position)
            {
              return offset;
            }

            break;
          }
          case SeekOrigin.Current:
          {
            if (offset == 0)
            {
              return resource.position;
            }

            break;
          }
        }

        throw new NotSupportedException();
      }

      /// <summary>
      /// No operation is supported.
      /// </summary>
      public override void SetLength(long value)
      {
        throw new NotSupportedException();
      }

      /// <summary>
      /// Writes data into a cache.
      /// </summary>
      /// <param name="buffer">A buffer to write.</param>
      /// <param name="offset">An offset within the buffer.</param>
      /// <param name="count">A number of bytes to write.</param>
      public override void Write(byte[] buffer, int offset, int count)
      {
        CheckNotDisposed();
        resource.Write(buffer, offset, count);
      }

      /// <summary>
      /// Disposes this instance.
      /// </summary>
      /// <param name="disposing">A dispose indicator.</param>
      protected override void Dispose(bool disposing)
      {
        try
        {
          disposed = true;
          resource.Eof();
        }
        finally
        {
          base.Dispose(disposing);
        }
      }

      /// <summary>
      /// Checks that OutputSourceStream instance is not disposed.
      /// </summary>
      private void CheckNotDisposed()
      {
        if (disposed)
        {
          throw new ObjectDisposedException("OutputSourceStream");
        }
      }

      /// <summary>
      /// A disposed indicator.
      /// </summary>
      private volatile bool disposed;

      /// <summary>
      /// A stream resource instance.
      /// </summary>
      private readonly StreamResource resource;
    }

    /// <summary>
    /// A disposed indicator.
    /// </summary>
    private int disposed;

    /// <summary>
    /// A use count.
    /// </summary>
    private int useCount;

    /// <summary>
    /// A cache stream.
    /// </summary>
    private FileStream cache;

    /// <summary>
    /// An input wait event.
    /// </summary>
    private readonly ManualResetEventSlim inputEvent = 
      new ManualResetEventSlim();

    /// <summary>
    /// A time spent reading data from source.
    /// </summary>
    private readonly Stopwatch readWatch = new Stopwatch();

    /// <summary>
    /// A time spend consuming read data.
    /// </summary>
    private readonly Stopwatch consumeWatch = new Stopwatch();

    /// <summary>
    /// Input stream position.
    /// </summary>
    private long position;

    /// <summary>
    /// Indicates whether end of input is reached.
    /// </summary>
    private volatile bool eof;

    /// <summary>
    /// A task that is completed when cached stream is flashed.
    /// </summary>
    private TaskCompletionSource<bool> flushCompletion;

    /// <summary>
    /// Settings instance.
    /// </summary>
    private readonly Settings settings;
  }
}
