﻿namespace Web
{
  using System;
  using System.Collections.Generic;
  using System.Linq;
  using System.Runtime.Serialization;
  using System.ServiceModel;
  using System.ServiceModel.Activation;
  using System.ServiceModel.Web;
  using System.Text;
  using System.IO;
  using System.Xml;
  using System.Xml.Xsl;
  using System.Web;
  using System.Web.Caching;
  using NesterovskyBros.DataStream;
  using NesterovskyBros.Xml;

  [ServiceContract(
    Namespace = "http://www.nesterovsky-bros.com",
    SessionMode = SessionMode.Allowed)]
  [AspNetCompatibilityRequirements(
    RequirementsMode = AspNetCompatibilityRequirementsMode.Allowed)]
  public class Service
  {
    [OperationContract]
    [WebGet(RequestFormat = WebMessageFormat.Json)]
    public Stream GetPeopleHtml(int count, int seed)
    {
      var context = WebOperationContext.Current;

      context.OutgoingResponse.ContentType = "text/html";
      context.OutgoingResponse.Headers["Content-Disposition"] = 
        "attachment;filename=reports.html";

      var transform = Extensions.GetStylesheet("~/People.xslt");

      return Extensions.GetPipedStream(
        output =>
        {
          // We have a streamed business data.
          var people = Data.CreateRandomData(count, seed, 0, count);

          // We want to see it as streamed xml data.
          using (var stream =
            people.ToXmlStream("people", "http://www.nesterovsky-bros.com"))
          using (var reader = XmlReader.Create(stream))
          {
            // XPath forward navigator is used as an input source.
            transform.Transform(
              new ForwardXPathNavigator(reader),
              new XsltArgumentList(),
              output);
          }
        });
    }

    /// <summary>
    /// Generates Excel according with request's parameters.
    /// </summary>
    /// <returns>Excel stream.</returns>
    [OperationContract]
    [WebGet(
      RequestFormat = WebMessageFormat.Json,
      BodyStyle = WebMessageBodyStyle.WrappedRequest)]
    public Stream GetExcel(
      int count, 
      int seed, 
      string chartType, 
      string sortBy)
    {
      // Emulate request data and cache it.  
      var data = Extensions.ToBufferedEnumerable(
            () => Data.CreateRandomData(count, seed, 0, count),
            1000
          );

      // note: it's possible that the requested template 
      //       consumes more than one data source. For example,
      //       instead of grouping data in XSLT you may prepare
      //       groupped data sources in code and just pass them
      //       as inputs.
      var inputs = new Dictionary<string, Func<Stream>>()
      {
        // this way we create a stream of business data
        { 
          "data", 
          () => data.ToXmlStream("people", "http://www.nesterovsky-bros.com") 
        }
      };

      // these are arguments for XSLT transformations
      var args = new Dictionary<string, object>
      {
        { "sortBy", sortBy }
      };

      // generate XLSX as a file attachment
      return ExportToExcel(chartType, inputs, args);
    }

    /// <summary>
    /// Generates a XLSX as a file attachment.
    /// </summary>
    /// <param name="name">a template name to export.</param>
    /// <param name="inputs">a dictionary with named input data.</param>
    /// <returns>an output stream with excel content.</returns>
    protected static Stream ExportToExcel(
      string name,
      IDictionary<string, Func<Stream>> inputs = null,
      IDictionary<string, object> args = null)
    {
      var context = WebOperationContext.Current;

      context.OutgoingResponse.ContentType = "application/vnd.ms-excel";
      context.OutgoingResponse.Headers["Content-Disposition"] =
        "attachment; filename=" + name + ".xlsx";

      // resolves a template location by its name
      var basePath =
        Path.Combine(HttpContext.Current.Server.MapPath("~/Excel"), name);

      if (inputs == null)
      {
        inputs = new Dictionary<string, Func<Stream>>();
      }

      return Extensions.GetPipedStream(
        stream =>
          Extensions.CreateXslxFile(
            stream,              // an output stream
            basePath,            // a base path for the requested template 
            id => inputs[id](),  // a factory for input streams
            args));              // XSLT input arguments
    }
  }
}
